import numpy as np
import scipy.sparse as sp
import os

from sklearn.decomposition import PCA
from sklearn.cluster import KMeans
from sklearn.neighbors import kneighbors_graph
from sklearn.metrics import adjusted_mutual_info_score,normalized_mutual_info_score,adjusted_rand_score,silhouette_score,davies_bouldin_score,calinski_harabasz_score

import torch
import torch.nn as nn
from torch.nn import Parameter
import torch.nn.functional as F
import torch.optim as optim
from torch_sparse import SparseTensor
from torch_geometric.utils import add_self_loops,remove_self_loops,to_undirected
from torch_geometric.data import Data
from torch_geometric.loader import DataLoader

from scMAGCA.loss import *
from scMAGCA.layers import *
from scMAGCA.utils import *
from scMAGCA.Dataset import MyDataset

def dopca(X, dim=50):
    pcaten = PCA(n_components=dim,random_state=42)
    X_pca = pcaten.fit_transform(X)
    return X_pca

def get_adj(count, k=30, pca=100):
    if pca:
        countp = dopca(count, dim=pca)
    else:
        countp = count
    A = kneighbors_graph(countp, k, metric="cosine", include_self=False).toarray()
    adj = sp.coo_matrix(A)
    return adj  

class scMultiCluster(nn.Module):
    def __init__(self, input_dim1, input_dim2, alpha, beta, gama, device="cuda"):
        super(scMultiCluster, self).__init__()
        self.input_dim1 = input_dim1
        self.input_dim2 = input_dim2
        self.alpha = alpha
        self.beta = beta
        self.gama = gama
        self.device = device
        self.z_dim = 32
        # Graph Encoder
        self.encoder = Encoder(layer_config=[input_dim1+input_dim2,1024,256,64,self.z_dim])
        # Linear Decoder
        self.decoder = nn.Sequential(
            nn.Linear(in_features=self.z_dim, out_features=512),nn.BatchNorm1d(512),nn.PReLU(),
            nn.Linear(in_features=512, out_features=1024),nn.BatchNorm1d(1024),nn.PReLU(),
            nn.Linear(in_features=1024, out_features=input_dim1+input_dim2),
        ).to(self.device)
        self.dec_mean = nn.Sequential(nn.Linear(self.z_dim, 256),nn.Linear(256, 512),nn.Linear(512,input_dim1+input_dim2), MeanAct())
        self.dec_disp = nn.Sequential(nn.Linear(self.z_dim, 256),nn.Linear(256, 512),nn.Linear(512,input_dim1+input_dim2), DispAct())
        self.dec_pi = nn.Sequential(nn.Linear(self.z_dim, 256),nn.Linear(256, 512),nn.Linear(512,input_dim1+input_dim2), nn.Sigmoid())
        self.zinb_loss = ZINBLoss()

    def pretrain_autoencoder(self, X1, X2, X1_raw, X2_raw, epochs=400, file=None, ad_out=16):
        print("Pretraining stage")
        self.num = X1.shape[0]
        X1, X2 = torch.tensor(X1), torch.tensor(X2)
        self.X = torch.cat([X1, X2], dim=-1)
        X1_raw, X2_raw = torch.tensor(X1_raw), torch.tensor(X2_raw)
        self.X_raw1, self.X_raw2 = X1_raw.to(self.device), X2_raw.to(self.device)
        self.X_raw = torch.cat([X1_raw, X2_raw,], dim=-1).to(self.device)

        # 图邻接矩阵A(细胞连接图)
        adj = get_adj(self.X)
        edge_index = torch.tensor(np.array([adj.row, adj.col]), dtype=torch.long)
        edge_index, _ = remove_self_loops(edge_index)
        edge_index = to_undirected(edge_index, self.num)
        x = torch.tensor(self.X, dtype=torch.float32)
        y = None
        data = Data(x=x, edge_index=edge_index, y=y)
        data.edge_attr = torch.ones(data.edge_index.shape[1])
        nodes = torch.tensor(np.arange(data.num_nodes), dtype=torch.long)
        edge_index, edge_attr = add_self_loops(data.edge_index, data.edge_attr)
        data = Data(nodes=nodes, edge_index=data.edge_index, edge_attr=data.edge_attr, x=data.x, y=data.y,
                    num_nodes=data.num_nodes, neighbor_index=edge_index, neighbor_attr=edge_attr)

        data.to(self.device)
        if not os.path.exists('../datasets/'+file+'/raw'):
            os.makedirs('../datasets/'+file+'/raw')
        if not os.path.exists('../datasets/'+file+'/processed'):
            os.makedirs('../datasets/'+file+'/processed')       
        torch.save(data, '../datasets/'+file+'/raw/raw.pt')
        
        dataset = MyDataset(file=file)
        dataloader = DataLoader(dataset, batch_size=128, shuffle=True)
        optimizer = optim.Adam(filter(lambda p: p.requires_grad, self.parameters()), lr=1e-3, amsgrad=True)

        self.discriminator = Discriminator(self.z_dim,ad_out).to(self.device) #CITE-seq:16; SMAGE-seq:32
        self.real_distribution = torch.randn(self.num, self.z_dim).to(self.device)

        for epoch in range(epochs):
            for batch_data in dataloader:
                z = self.encoder(x=batch_data.x, 
                                edge_index=SparseTensor(row=batch_data.edge_index[0], col=batch_data.edge_index[1]), 
                                edge_weight=batch_data.edge_attr).to(self.device)

                # reconstruct_loss
                recon_x = self.decoder(z)
                recon_loss = F.mse_loss(recon_x, batch_data.x)

                # ZINB loss
                mean,disp,pi = self.dec_mean(z),self.dec_disp(z),self.dec_pi(z)
                zinb_loss = self.zinb_loss(x=self.X_raw, mean=mean, disp=disp, pi=pi)

                # adversarial loss
                self.dis_real_logit = self.discriminator(self.real_distribution)
                self.dis_fake_logit = self.discriminator(z)
                discriminator_loss = adversarial_loss(self.dis_real_logit,self.dis_fake_logit)

                loss = self.alpha*recon_loss + self.beta*zinb_loss + self.gama*discriminator_loss

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

            print('Pretrain epoch {}, recon_loss:{:.6f}, zinb_loss:{:.6f}, adversial_loss:{:.6f}'.format(epoch+1, recon_loss.item(), zinb_loss.item(), discriminator_loss.item()))

        data.to(self.device)
        return z

    def fit(self, y=None, n_clusters=-1, num_epochs=400, file=None, tol=1e-3, pretrain_latent=None, resolution=.08):
        print("Clustering stage")
        dataset = MyDataset(file=file)
        dataloader = DataLoader(dataset, batch_size=128, shuffle=True)
        optimizer = optim.Adadelta(filter(lambda p: p.requires_grad, self.parameters()), lr=1, rho=.95)

        if y is not None:
            print("Initializing cluster centers with kmeans.")
            kmeans = KMeans(n_clusters=n_clusters,n_init=20)
            self.mu = Parameter(torch.Tensor(n_clusters, self.z_dim), requires_grad=True).to(self.device)
            for _, batch_data in enumerate(dataloader):
                z = self.encoder(x=batch_data.x, edge_index=SparseTensor(row=batch_data.edge_index[0], col=batch_data.edge_index[1]), edge_weight=batch_data.edge_attr)
            self.y_pred = kmeans.fit_predict(z.data.cpu().numpy())
            self.y_pred_last = self.y_pred
            self.mu.data.copy_(torch.Tensor(kmeans.cluster_centers_))
            
            ami = np.round(adjusted_mutual_info_score(y, self.y_pred), 5)
            nmi = np.round(normalized_mutual_info_score(y, self.y_pred), 5)
            ari = np.round(adjusted_rand_score(y, self.y_pred), 5)
            acc = np.round(cluster_acc(y, self.y_pred), 5)

            print('Initializing k-means: AMI= %.4f, NMI= %.4f, ARI= %.4f, ACC= %.4f' % (ami,nmi,ari,acc))
            self.train()
        
            final_nmi, final_ami, final_ari, final_acc= 0, 0, 0, 0
            for epoch in range(num_epochs):
                for _, batch_data in enumerate(dataloader):
                    z = self.encoder(x=batch_data.x, 
                                    edge_index=SparseTensor(row=batch_data.edge_index[0], col=batch_data.edge_index[1]), 
                                    edge_weight=batch_data.edge_attr)
                  
                    # reconstruct loss
                    recon_x = self.decoder(z)
                    recon_loss = F.mse_loss(recon_x, batch_data.x)

                    # zinb loss
                    mean,disp,pi = self.dec_mean(z),self.dec_disp(z),self.dec_pi(z)
                    zinb_loss = self.zinb_loss(x=self.X_raw, mean=mean, disp=disp, pi=pi)

                    # cluster loss
                    q = soft_assign(z, self.mu)
                    p = target_distribution(q).data
                    self.y_pred = torch.argmax(q, dim=1).data.cpu().numpy()
                    cluste_loss = cluster_loss(p,q)

                    # overall loss
                    loss = self.alpha*recon_loss + self.beta*zinb_loss + cluste_loss

                    optimizer.zero_grad()
                    loss.backward()
                    torch.nn.utils.clip_grad_norm_(self.mu, 1)
                    optimizer.step()

                self.y_pred = torch.argmax(q, dim=1).data.cpu().numpy()
                final_ami = ami = np.round(adjusted_mutual_info_score(y, self.y_pred), 5)
                final_nmi = nmi = np.round(normalized_mutual_info_score(y, self.y_pred), 5)
                final_ari = ari = np.round(adjusted_rand_score(y, self.y_pred), 5)
                final_acc = acc = np.round(cluster_acc(y, self.y_pred), 5)

                print('Training epoch {}, recon_loss:{:.6f}, zinb_loss:{:.6f}, cluster_loss:{:.6f}'.format(epoch+1,recon_loss.item(),zinb_loss.item(),cluste_loss.item()))
                print('Clustering   %d: AMI= %.4f, NMI= %.4f, ARI= %.4f, ACC= %.4f' % (epoch+1,ami,nmi,ari,acc))
                     
                delta_label = np.sum(self.y_pred != self.y_pred_last).astype(np.float32) / self.num
                print(delta_label)
                self.y_pred_last, self.embedding = self.y_pred, z
                if epoch>0 and delta_label < tol:
                    print('delta_label ', delta_label, '< tol ', tol)
                    print("Reach tolerance threshold. Stopping training.")
                    break

            self.y_pred_last, self.embedding = self.y_pred, z
            print('Final Result : AMI= %.4f, NMI= %.4f, ARI= %.4f, ACC= %.4f'% (final_ami,final_nmi,final_ari,final_acc))         
                
            return self.y_pred_last, self.embedding

        else:
            n_clusters = GetCluster(pretrain_latent.cpu().detach().numpy(),n=20,res=resolution)
            print("Initializing cluster centers with kmeans.")
            kmeans = KMeans(n_clusters=n_clusters,n_init=20)
            self.mu = Parameter(torch.Tensor(n_clusters, self.z_dim), requires_grad=True).to(self.device)

            for _, batch_data in enumerate(dataloader):
                z = self.encoder(x=batch_data.x, edge_index=SparseTensor(row=batch_data.edge_index[0], col=batch_data.edge_index[1]), edge_weight=batch_data.edge_attr)
            self.y_pred = kmeans.fit_predict(z.data.cpu().numpy())
            self.y_pred_last = self.y_pred
            self.mu.data.copy_(torch.Tensor(kmeans.cluster_centers_))
            asw = np.round(silhouette_score(z.data.cpu().numpy(), self.y_pred), 5)
            db = np.round(davies_bouldin_score(z.data.cpu().numpy(), self.y_pred), 5)
            ch = np.round(calinski_harabasz_score(z.data.cpu().numpy(), self.y_pred), 5)

            print('Initializing k-means: ASW= %.4f, DB= %.4f, CH= %.4f' % (asw,db,ch))
            self.train()
        
            final_asw, final_db, final_ch = 0, 0, 0
            for epoch in range(num_epochs):
                for _, batch_data in enumerate(dataloader):
                    z = self.encoder(x=batch_data.x, 
                                    edge_index=SparseTensor(row=batch_data.edge_index[0], col=batch_data.edge_index[1]), 
                                    edge_weight=batch_data.edge_attr)
                  
                    # reconstruct loss
                    recon_x = self.decoder(z)
                    recon_loss = F.mse_loss(recon_x, batch_data.x)

                    # zinb loss
                    mean,disp,pi = self.dec_mean(z),self.dec_disp(z),self.dec_pi(z)
                    zinb_loss = self.zinb_loss(x=self.X_raw, mean=mean, disp=disp, pi=pi)
                    
                    # cluster loss
                    q = soft_assign(z, self.mu)
                    p = target_distribution(q).data
                    self.y_pred = torch.argmax(q, dim=1).data.cpu().numpy()
                    cluste_loss = cluster_loss(p,q)

                    # overall loss
                    loss = self.alpha*recon_loss + self.beta*zinb_loss + cluste_loss

                    optimizer.zero_grad()
                    loss.backward()
                    torch.nn.utils.clip_grad_norm_(self.mu, 1)
                    optimizer.step()
            
                self.y_pred = torch.argmax(q, dim=1).data.cpu().numpy()
                final_asw = asw = np.round(silhouette_score(z.data.cpu().numpy(), self.y_pred), 5)
                final_db = db = np.round(davies_bouldin_score(z.data.cpu().numpy(), self.y_pred), 5)
                final_ch = ch = np.round(calinski_harabasz_score(z.data.cpu().numpy(), self.y_pred), 5)

                print('Training epoch {}, recon_loss:{:.6f}, zinb_loss:{:.6f}, cluster_loss:{:.6f}'.format(epoch+1,recon_loss.item(),zinb_loss.item(),cluste_loss.item()))
                print('Clustering   %d: ASW= %.4f, DB= %.4f, CH= %.4f' % (epoch+1,asw,db,ch))

            self.y_pred_last, self.embedding = self.y_pred, z
            print('Final Result : ASW= %.4f, DB= %.4f, CH= %.4f'% (final_asw,final_db,final_ch))  

            return self.y_pred_last, self.embedding


import torch
import torch.nn as nn
import torch.nn.functional as F

tau = 1
alpha = 1
EPS = 1e-15

# NB Loss
class NBLoss(nn.Module):
    def __init__(self):
        super(NBLoss, self).__init__()

    def forward(self, x, mean, disp, scale_factor=1.0):
        eps = 1e-10
        #scale_factor = scale_factor[:, None]
        mean = mean * scale_factor
        
        t1 = torch.lgamma(disp+eps) + torch.lgamma(x+1.0) - torch.lgamma(x+disp+eps)
        t2 = (disp+x) * torch.log(1.0 + (mean/(disp+eps))) + (x * (torch.log(disp+eps) - torch.log(mean+eps)))
        result = t1 + t2

        result = torch.mean(result)
        return result

# ZINB Loss
class ZINBLoss(nn.Module):
    def __init__(self):
        super(ZINBLoss, self).__init__()

    def forward(self, x, mean, disp, pi, scale_factor=1.0, ridge_lambda=0.0):
        eps = 1e-10
        #scale_factor = scale_factor[:, None]
        mean = mean * scale_factor

        t1 = torch.lgamma(disp + eps) + torch.lgamma(x + 1.0) - torch.lgamma(x + disp + eps)
        t2 = (disp + x) * torch.log(1.0 + (mean / (disp + eps))) + (x * (torch.log(disp + eps) - torch.log(mean + eps)))
        nb_final = t1 + t2

        nb_case = nb_final - torch.log(1.0 - pi + eps)
        zero_nb = torch.pow(disp / (disp + mean + eps), disp)
        zero_case = -torch.log(pi + ((1.0 - pi) * zero_nb) + eps)
        result = torch.where(torch.le(x, 1e-8), zero_case, nb_case)

        if ridge_lambda > 0:
            ridge = ridge_lambda * torch.square(pi)
            result += ridge

        result = torch.mean(result)
        return result   

def soft_assign(z, mu):
    q = 1.0 / (1.0 + torch.sum((z.unsqueeze(1) - mu) ** 2, dim=2) / alpha)
    q = q ** ((alpha + 1.0) / 2.0)
    q = (q.t() / torch.sum(q, dim=1)).t()
    return q

def target_distribution(q):
    p = q ** 2 / q.sum(0)
    return (p.t() / p.sum(1)).t() 

# Cluster Loss
def cluster_loss(p, q):
    def kld(target, pred):
        return torch.mean(torch.sum(target * torch.log(target / (pred + 1e-6)), dim=-1))
    kldloss = kld(p, q)
    return kldloss
    
# Adversarial Loss
def adversarial_loss(real_output, fake_output):
    real_loss = F.binary_cross_entropy(real_output, torch.ones_like(real_output))
    fake_loss = F.binary_cross_entropy(fake_output, torch.zeros_like(fake_output))
    total_loss = real_loss + fake_loss
    return total_loss


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import scipy as sp
import pandas as pd
import scanpy as sc
from sklearn.model_selection import train_test_split

def read_dataset(adata, transpose=False, test_split=False, copy=False):

    if isinstance(adata, sc.AnnData):
        if copy:
            adata = adata.copy()
    elif isinstance(adata, str):
        adata = sc.read(adata)
    else:
        raise NotImplementedError

    norm_error = 'Make sure that the dataset (adata.X) contains unnormalized count data.'
    assert 'n_count' not in adata.obs, norm_error

    if adata.X.size < 50e6: # check if adata.X is integer only if array is small
        if sp.sparse.issparse(adata.X):
            assert (adata.X.astype(int) != adata.X).nnz == 0, norm_error
        else:
            assert np.all(adata.X.astype(int) == adata.X), norm_error

    if transpose: adata = adata.transpose()

    if test_split:
        train_idx, test_idx = train_test_split(np.arange(adata.n_obs), test_size=0.1, random_state=42)
        spl = pd.Series(['train'] * adata.n_obs)
        spl.iloc[test_idx] = 'test'
        adata.obs['DCA_split'] = spl.values
    else:
        adata.obs['DCA_split'] = 'train'

    adata.obs['DCA_split'] = adata.obs['DCA_split'].astype('category')
    print('### Autoencoder: Successfully preprocessed {} features and {} cells.'.format(adata.n_vars, adata.n_obs))

    return adata

def preprocess_dataset(adata, size_factors=False, normalize_input=True, logtrans_input=True):
    def seurat_clr(x):
    # TODO: support sparseness
        s = np.sum(np.log1p(x[x > 0]))
        exp = np.exp(s / len(x))
        return np.log1p(x / exp)

    adata.raw = adata.copy()
    sc.pp.normalize_total(adata)
    
    if size_factors:
        adata.obs['size_factors'] = adata.obs.n_counts / np.median(adata.obs.n_counts)
    else:
        adata.obs['size_factors'] = 1.0

    # apply to dense or sparse matrix, along axis. returns dense matrix
    adata.X = np.apply_along_axis(
        seurat_clr, 1, (adata.raw.X.A if sp.sparse.issparse(adata.raw.X) else adata.raw.X)
    )

    if logtrans_input:
        sc.pp.log1p(adata)

    if normalize_input:
        sc.pp.scale(adata)

    return adata 


from munkres import Munkres
import scanpy as sc
from scipy import sparse
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.utils.linear_assignment_ import linear_assignment

def cluster_acc(y_true, y_pred):
    """
    Calculate clustering accuracy. Require scikit-learn installed
    # Arguments
        y: true labels, numpy.array with shape `(n_samples,)`
        y_pred: predicted labels, numpy.array with shape `(n_samples,)`
    # Return
        accuracy, in [0,1]
    """
    y_true = y_true.astype(np.int64)
    assert y_pred.size == y_true.size
    D = max(y_pred.max(), y_true.max()) + 1
    w = np.zeros((D, D), dtype=np.int64)
    for i in range(y_pred.size):
        w[y_pred[i], y_true[i]] += 1 # 创建了一个标记矩阵
    
    ind = linear_assignment(w.max() - w) # 匈牙利算法
    return sum([w[i, j] for i, j in ind]) * 1.0 / y_pred.size

# 当数据是无标签时，会估计k值
def GetCluster(X, res, n):
    adata0=sc.AnnData(X)
    if adata0.shape[0]>200000: # shape(细胞数, 基因数)
       np.random.seed(adata0.shape[0])#set seed 
       adata0=adata0[np.random.choice(adata0.shape[0],200000,replace=False)] 
    sc.pp.neighbors(adata0, n_neighbors=n, use_rep="X")
    sc.tl.louvain(adata0,resolution=res)
    Y_pred_init=adata0.obs['louvain']
    Y_pred_init=np.asarray(Y_pred_init,dtype=int)
    if np.unique(Y_pred_init).shape[0]<=1:
        #avoid only a cluster
        exit("Error: There is only a cluster detected. The resolution:"+str(res)+"is too small, please choose a larger resolution!!")
    else: 
        print("Estimated n_clusters is: ", np.shape(np.unique(Y_pred_init))[0]) 
    return(np.shape(np.unique(Y_pred_init))[0])
    
def best_map(L1,L2):
            #L1 should be the groundtruth labels and L2 should be the clustering labels we got
            Label1 = np.unique(L1)
            nClass1 = len(Label1)
            Label2 = np.unique(L2)
            nClass2 = len(Label2)
            nClass = np.maximum(nClass1,nClass2)
            G = np.zeros((nClass,nClass))
            for i in range(nClass1):
                ind_cla1 = L1 == Label1[i]
                ind_cla1 = ind_cla1.astype(float)
                for j in range(nClass2):
                    ind_cla2 = L2 == Label2[j]
                    ind_cla2 = ind_cla2.astype(float)
                    G[i,j] = np.sum(ind_cla2 * ind_cla1)
            m = Munkres()
            index = m.compute(-G.T)
            index = np.array(index)
            c = index[:,1]
            newL2 = np.zeros(L2.shape)
            for i in range(nClass2):
                newL2[L2 == Label2[i]] = Label1[c[i]]
            return newL2
 
def geneSelection(data, threshold=0, atleast=10, 
                  yoffset=.02, xoffset=5, decay=1.5, n=None, 
                  plot=True, markers=None, genes=None, figsize=(6,3.5),
                  markeroffsets=None, labelsize=10, alpha=1, verbose=1):
    
    if sparse.issparse(data):
        zeroRate = 1 - np.squeeze(np.array((data>threshold).mean(axis=0)))
        A = data.multiply(data>threshold)
        A.data = np.log2(A.data)
        meanExpr = np.zeros_like(zeroRate) * np.nan
        detected = zeroRate < 1
        meanExpr[detected] = np.squeeze(np.array(A[:,detected].mean(axis=0))) / (1-zeroRate[detected])
    else:
        zeroRate = 1 - np.mean(data>threshold, axis=0)
        meanExpr = np.zeros_like(zeroRate) * np.nan
        detected = zeroRate < 1
        mask = data[:,detected]>threshold
        logs = np.zeros_like(data[:,detected]) * np.nan
        logs[mask] = np.log2(data[:,detected][mask])
        meanExpr[detected] = np.nanmean(logs, axis=0)

    lowDetection = np.array(np.sum(data>threshold, axis=0)).squeeze() < atleast
    zeroRate[lowDetection] = np.nan
    meanExpr[lowDetection] = np.nan
            
    if n is not None:
        up = 10
        low = 0
        for t in range(100):
            nonan = ~np.isnan(zeroRate)
            selected = np.zeros_like(zeroRate).astype(bool)
            selected[nonan] = zeroRate[nonan] > np.exp(-decay*(meanExpr[nonan] - xoffset)) + yoffset
            if np.sum(selected) == n:
                break
            elif np.sum(selected) < n:
                up = xoffset
                xoffset = (xoffset + low)/2
            else:
                low = xoffset
                xoffset = (xoffset + up)/2
        if verbose>0:
            print('Chosen offset: {:.2f}'.format(xoffset))
    else:
        nonan = ~np.isnan(zeroRate)
        selected = np.zeros_like(zeroRate).astype(bool)
        selected[nonan] = zeroRate[nonan] > np.exp(-decay*(meanExpr[nonan] - xoffset)) + yoffset
                
    if plot:
        if figsize is not None:
            plt.figure(figsize=figsize)
        plt.ylim([0, 1])
        if threshold>0:
            plt.xlim([np.log2(threshold), np.ceil(np.nanmax(meanExpr))])
        else:
            plt.xlim([0, np.ceil(np.nanmax(meanExpr))])
        x = np.arange(plt.xlim()[0], plt.xlim()[1]+.1,.1)
        y = np.exp(-decay*(x - xoffset)) + yoffset
        if decay==1:
            plt.text(.4, 0.2, '{} genes selected\ny = exp(-x+{:.2f})+{:.2f}'.format(np.sum(selected),xoffset, yoffset), 
                     color='k', fontsize=labelsize, transform=plt.gca().transAxes)
        else:
            plt.text(.4, 0.2, '{} genes selected\ny = exp(-{:.1f}*(x-{:.2f}))+{:.2f}'.format(np.sum(selected),decay,xoffset, yoffset), 
                     color='k', fontsize=labelsize, transform=plt.gca().transAxes)

        plt.plot(x, y, color=sns.color_palette()[1], linewidth=2)
        xy = np.concatenate((np.concatenate((x[:,None],y[:,None]),axis=1), np.array([[plt.xlim()[1], 1]])))
        t = plt.matplotlib.patches.Polygon(xy, color=sns.color_palette()[1], alpha=.4)
        plt.gca().add_patch(t)
        
        plt.scatter(meanExpr, zeroRate, s=1, alpha=alpha, rasterized=True)
        if threshold==0:
            plt.xlabel('Mean log2 nonzero expression')
            plt.ylabel('Frequency of zero expression')
        else:
            plt.xlabel('Mean log2 nonzero expression')
            plt.ylabel('Frequency of near-zero expression')
        plt.tight_layout()
        
        if markers is not None and genes is not None:
            if markeroffsets is None:
                markeroffsets = [(0, 0) for g in markers]
            for num,g in enumerate(markers):
                i = np.where(genes==g)[0]
                plt.scatter(meanExpr[i], zeroRate[i], s=10, color='k')
                dx, dy = markeroffsets[num]
                plt.text(meanExpr[i]+dx+.1, zeroRate[i]+dy, g, color='k', fontsize=labelsize)

    return selected



import numpy as np
import pandas as pd
from time import time
import torch
import scanpy as sc
import random

from sklearn.preprocessing import OneHotEncoder
from scMAGCA.preprocess import read_dataset, preprocess_dataset
from scMAGCA.utils import *
from scMAGCA.scMAGCA_batch import scMultiCluster

if __name__ == "__main__":
    # setting the hyper parameters
    import argparse
    parser = argparse.ArgumentParser(description='train',formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--n_clusters', type=int)
    parser.add_argument('--n_batches', type=int)
    parser.add_argument('--data_file', required=True)
    parser.add_argument('--pretrain_epochs', default=400, type=int)
    parser.add_argument('--maxiter', default=2000, type=int)
    parser.add_argument('--save_dir', default='../results/')
    parser.add_argument('--embedding_file', action='store_true', default=True)
    parser.add_argument('--prediction_file', action='store_true', default=True)
    parser.add_argument('--ad_out', type=int, default=16, help='Output dim of discriminator')
    parser.add_argument('--tol', default=0.001, type=float, help='Criterion to stop the model')
    parser.add_argument('--resolution', default=0.08, type=float, help='Resolution parameter to estimate K')
    parser.add_argument('--filter1', action='store_true', default=False, help='Do ADT/ATAC selection')
    parser.add_argument('--filter2', action='store_true', default=False, help='Do mRNA selection')
    parser.add_argument('--f1', default=2000, type=int, help='Number of ADT/ATAC after feature selection')
    parser.add_argument('--f2', default=2000, type=int, help='Number of mRNA after feature selection')
    parser.add_argument('--alpha', default=0.2, type=float, help='Weight of reconstruction loss')
    parser.add_argument('--beta', default=0.8, type=float, help='Weight of ZINB loss')
    parser.add_argument('--gama', default=0.01, type=float, help='Weight of adversarial loss')
    parser.add_argument('--device', default='cuda')
    args = parser.parse_args()
    print(args)

    # set seed
    random.seed(3407)
    np.random.seed(3407)
    torch.manual_seed(3407)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.enabled = False
    torch.backends.cudnn.benchmark = False

    ### read multiBatch dataset
    x1 = np.array(sc.read_h5ad('../datasets/GSE164378/GSE164378_adt.h5ad').to_df()).astype('float32')
    x2 = np.array(sc.read_h5ad('../datasets/GSE164378/GSE164378_rna.h5ad').to_df()).astype('float32')
    y = np.array(pd.read_csv('../datasets/GSE164378/GSE164378_label.csv')['Cluster']).astype('float32')
    b = np.array(pd.read_csv('../datasets/GSE164378/GSE164378_batch.csv')['Batch']).astype('float32')
    enc = OneHotEncoder()
    enc.fit(b.reshape(-1, 1))
    B = enc.transform(b.reshape(-1, 1)).toarray()

    # Gene filter
    if args.filter1:
        importantGenes = geneSelection(x1, n=args.f1)
        x1 = x1[:, importantGenes]
    if args.filter2:
        importantGenes = geneSelection(x2, n=args.f2)
        x2 = x2[:, importantGenes]
        
    # preprocessing scRNA-seq read counts matrix
    adata1 = sc.AnnData(x1)
    adata1 = read_dataset(adata1, copy=True)
    adata1 = preprocess_dataset(adata1, normalize_input=True, logtrans_input=True)
    print(adata1)

    adata2 = sc.AnnData(x2)
    adata2 = read_dataset(adata2, copy=True)
    adata2 = preprocess_dataset(adata2, normalize_input=True, logtrans_input=True)
    print(adata2)
    
    input_size1 = adata1.n_vars
    input_size2 = adata2.n_vars

    model = scMultiCluster(input_dim1=input_size1,input_dim2=input_size2,n_batch=args.n_batches, 
                           alpha=args.alpha,beta=args.beta,gama=args.gama,device=args.device).to(args.device)
    print(str(model))

    t0 = time()
    pretrain_latent = model.pretrain_autoencoder(
                        X1=adata1.X, X2=adata2.X, X1_raw=adata1.raw.X, X2_raw=adata2.raw.X, B=B,
                        epochs=args.pretrain_epochs, file=args.data_file)
    print('Pretraining time: %d seconds.' % int(time() - t0))
 
    y_pred, final_latent = model.fit(y=y, n_clusters=args.n_clusters, num_epochs=args.maxiter, file=args.data_file)
    if args.prediction_file:
        y_pred_ = best_map(y, y_pred)
        y_pred_ = y_pred_.astype(int)
        np.savetxt(args.save_dir + str(args.data_file) + "_pred.csv", y_pred, delimiter=",")
    if args.embedding_file:
       final_latent = final_latent.cpu().detach().numpy()
       np.savetxt(args.save_dir + str(args.data_file) + "_embedding.csv", final_latent, delimiter=",")

    print('Total time: %d seconds.' % int(time() - t0))


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import GCNConv

class MeanAct(nn.Module):
    def __init__(self):
        super(MeanAct, self).__init__()

    def forward(self, x):
        return torch.clamp(torch.exp(x), min=1e-5, max=1e6)

class DispAct(nn.Module):
    def __init__(self):
        super(DispAct, self).__init__()

    def forward(self, x):
        return torch.clamp(F.softplus(x), min=1e-4, max=1e4)
    
class Encoder(nn.Module):
    def __init__(self, layer_config):
        super().__init__()
        self.stacked_gnn = nn.ModuleList(
            [GCNConv(layer_config[i - 1], layer_config[i]) for i in range(1, len(layer_config))])
        self.stacked_bns = nn.ModuleList(
            [nn.BatchNorm1d(layer_config[i], momentum=0.01) for i in range(1, len(layer_config))])
        self.stacked_prelus = nn.ModuleList([nn.PReLU() for _ in range(1, len(layer_config))])

    def forward(self, x, edge_index, edge_weight=None):
        for i, gnn in enumerate(self.stacked_gnn):
            x = gnn(x, edge_index, edge_weight)
            x = self.stacked_bns[i](x)
            x = self.stacked_prelus[i](x)

        return x
    
def buildNetwork(layers, activation="elu"):
    net = nn.Sequential()
    for i in range(1, len(layers)):
        net.add_module('linear%d'%i, nn.Linear(layers[i-1], layers[i]))
        net.add_module('batchnorm%d'%i, nn.BatchNorm1d(layers[i], affine=True))
        if activation=="relu":
            net.add_module('relu%d'%i, nn.ReLU())
        elif activation=="selu":
            net.add_module('selu%d'%i, nn.SELU())
        elif activation=="sigmoid":
            net.add_module('sigmoid%d'%i, nn.Sigmoid())
        elif activation=="elu":
            net.add_module('elu%d'%i, nn.ELU())
    return net

class Discriminator(nn.Module):
    def __init__(self, latent_dim, hidden_dim):
        super(Discriminator, self).__init__()
        self.fc1 = nn.Linear(latent_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, 1)
    
    def forward(self, x):
        x = F.relu(self.fc1(x))
        x = torch.sigmoid(self.fc2(x))
        return x


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="scMAGCA",
    version="1.0.10",
    author="zhouzeming",
    author_email="657608841@qq.com",
    description="scMAGCA, a multi-omics adversarial graph convolutional autoencoder framework to extract interpretable latent embeddings that characterize scMultiomics data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zemingzhou30/scMAGCA",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import os.path as osp
import torch
from torch_geometric.data import InMemoryDataset

class MyDataset(InMemoryDataset):
    """
    A PyTorch InMemoryDataset to build multi-view dataset through graph data augmentation
    """
    def __init__(self, root="../datasets/", file=None
                 , transform=None, pre_transform=None, pre_filter=None):
        super().__init__(root=root+file, transform=transform, pre_transform=pre_transform, pre_filter=pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[0])

    def process(self):
        print("Processing full batch data")
        data = torch.load(osp.join(self.raw_dir, self.raw_file_names[0]))
        data_list = [data]
        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]
        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]
        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[0])

    def download(self):
        pass

    @property
    def raw_dir(self) -> str:
        return osp.join(self.root, 'raw')
    
    @property
    def raw_file_names(self):
        return ['raw.pt']
    
    @property
    def processed_dir(self) -> str:
        return osp.join(self.root, 'processed')
    
    @property
    def processed_file_names(self):
        return ['processed.pt']